#include "mainwindow.h"
#include "ui_mainwindow.h"
#include <errno.h>

//Legacy constants, not used anymore
#define NOSAMPS 160
#define VOLTAGE 0.0048828125
#define FREQUENCY_STEP 1.781

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    ui->setupUi(this);

    settings = new Settings(this);
    connect(settings, SIGNAL(deviceSettings(int,int,int)), this, SLOT(sendDeviceSettings(int,int,int)));

    serial = new QSerialPort(this); //Construct serial object

    //Connect correspniindg serial object signals to their slots
    connect(serial, SIGNAL(readyRead()), this, SLOT(readData()));
    connect(ui->actionConnect, SIGNAL(triggered()), this, SLOT(openSerialPort()));
    connect(ui->actionDisconnect, SIGNAL(triggered()), this, SLOT(closeSerialPort()));
    connect(ui->actionSettings, SIGNAL(triggered()), this, SLOT(openSettings()));

    learner = new Learner(this);
    connect(learner, SIGNAL(gatherData(int,int,QString,QString)), this, SLOT(startGathering(int,int,QString,QString)));
    connect(ui->actionGather_data, SIGNAL(triggered()), this, SLOT(openLearnerDialog()));
    connect(ui->actionDo_Magic, SIGNAL(triggered(bool)), this, SLOT(enableMagic(bool)));

    terminal = new SerialTerminal(this);
    connect(ui->actionTerminal, SIGNAL(triggered()), this, SLOT(openTerminal()));
    connect(terminal,SIGNAL(sendToSerial(QByteArray)),this, SLOT(sendSerialData(QByteArray)));

    connect(ui->actionExit, SIGNAL(triggered()), this, SLOT(close()));

    numberOfDataStillNeeded = 0;
    ui->progressBar->setVisible(false);

    connect(ui->actionSimulate_Keyboard, SIGNAL(triggered()), this, SLOT(openSimulateKeyboard()));
    keyboardPlayground = new SimulateKeyboard(this);

    connect(ui->actionAbout, SIGNAL(triggered()), this, SLOT(aboutDialog()));

    numberOfSamples = NOSAMPS;

    dataASecondT = new QTimer(this);
    connect(dataASecondT, SIGNAL(timeout()), this, SLOT(dataASecond()));
    collected = 0;
}

void MainWindow::aboutDialog()
{
    QMessageBox::information(this, "About", "This is an interface application for xmega based board which is part of final year project called GD system. Developed by Raivis Strogonovs (625826).\n");
}

void MainWindow::sendDeviceSettings(int min, int max, int n)
{
    qDebug() << "Sending 2: " << QString("%1 %2 %3").arg(min).arg(max).arg(n).toLatin1();
    this->sendSerialData( "s" );
    this->sendSerialData( "c" );
    QByteArray configData = QByteArray( QString("%1 %2 %3\n").arg(min).arg(max).arg(n).toLatin1() );
    sendSerialData( configData );
    //this->sendSerialData( "r" );
    this->sendSerialData( "d" );
}

void MainWindow::openSimulateKeyboard()
{
    if(gestureNameList.size() == 0)
    {
        QMessageBox::critical(this, "Error", "Before using keyboard playground you must load model file.");
        return;
    }
    keyboardPlayground->show();
    keyboardPlayground->setGestures( gestureNameList );
}

void MainWindow::enableMagic(bool enable)
{
    //Is pattern recognition enabled?
    if(enable)
    {
        gestureNameList.clear();
        QString fileName = QFileDialog::getOpenFileName(this,
                tr("Open SVM model"), "",
                tr("SVM model data (*.model);;All Files (*)"));

        if (fileName.isEmpty())
        {
            ui->actionDo_Magic->blockSignals(true);
            ui->actionDo_Magic->setChecked(false);
            ui->actionDo_Magic->blockSignals(false);
                 return;
        }
        else
        {
            //QFile file(fileName);
            QDir dir;
            //qDebug() << dir.absoluteFilePath("gesture.svm.model");
            char buffer[1024];
            strcpy(buffer, fileName.toStdString().c_str() );

            // Load SVM model in its data structure
            if((model=svm_load_model(buffer))==0)
            {
                QMessageBox::critical(this, tr("Error"), tr("Can't open model file."));
                exit(1);
            }
            fileName.chop(5);
            QFile file( fileName.append("range") );
            if(!file.open(QIODevice::ReadOnly))
            {
                QMessageBox::critical(0, tr("Error"), file.errorString());
                return;
            }

            //Read the range, so we can perform scaling later
            QTextStream in(&file);
            in.readLine(); in.readLine();
            while(!in.atEnd())
            {
                QString line = in.readLine();
                QStringList fields = line.split(" ");

                int i = fields.at(0).toInt() - 1;
                double min = fields.at(1).toDouble();
                double max = fields.at(2).toDouble();

                rmin[i] = min;
                rmax[i] = max;

                //qDebug() << i << rmin[i] << rmax[i];
            }

            file.close();

            fileName.chop(5);
            //Read the names of gestures
            file.setFileName( fileName.append("name") );
            if(!file.open(QIODevice::ReadOnly))
            {
                QMessageBox::critical(0, tr("Error"), file.errorString());
                return;
            }

            in.setDevice(&file);
            while(!in.atEnd())
            {
                gestureNameList.push_back( in.readLine() );
            }
            file.close();
        }

        ui->statusBar->showMessage("SVM model loaded, ready to do some magic");
    }
}

void MainWindow::openTerminal()
{
    terminal->show();
}

void MainWindow::openSettings()
{
    settings->exec();    
}

void MainWindow::openLearnerDialog()
{
    learner->show();
}

void MainWindow::startGathering(int amount, int id, QString file, QString name)
{
    //initiates data gathering for learning
    numberOfDataStillNeeded = amount;
    numberOfDataNeeded = amount;
    ui->progressBar->setRange(0, amount);
    ui->progressBar->setVisible( true );
    this->id = id;
    this->file = file;
    this->gestureName = name;
    gatheredData.clear();
}

void MainWindow::doSomeMagic(QVector<double> samp)
{
    struct svm_node features[numberOfSamples+1];
    double *decVals;
    int x, noLabels, noDecVals;
    double r;



    noLabels = svm_get_nr_class( model );
    noDecVals=noLabels*(noLabels-1)/2;
    decVals=(double*)malloc(sizeof(double)*noDecVals);

    //Perform data scaling from -1 to 1
    for(x=0; x<numberOfSamples; x++)
    {
        features[x].index = x;

        r=(samp[x]-rmin[x+1]);
        r=r /(rmax[x+1]-rmin[x+1]);
        r=(r*2)-1;

        features[x+1].value = r;
    }

    features[x].index = -1;
    //Get the ID of closest data cluster
    r = svm_predict_values(model, features, decVals);
    writeToLogger( QString("Prediction value: %1 %2\n").arg(r) );

    /*switch((int)r)
    {
        case 1: ui->gestureLbl->setText("One finger"); break;
        case 2: ui->gestureLbl->setText("Four fingers"); break;
        case 4: ui->gestureLbl->setText("Two fingers"); break;
        case 3: ui->gestureLbl->setText("Hand near by"); break;
        default: ui->gestureLbl->setText("Nothing");
    }*/

    //Print the name of gesture
    if((int)r < gestureNameList.size())
        ui->gestureLbl->setText( gestureNameList.at( (int)r ) );

    //If key board simulator is open, emulate key press
    if(keyboardPlayground->isVisible())
        keyboardPlayground->gesture( (int)r );

}

void MainWindow::plot(QVector<double> values)
{
    ui->plot->clearGraphs();
    ui->plot->addGraph();
    ui->plot->graph(0)->setPen(QPen(Qt::blue)); // line color blue for first graph
    ui->plot->graph(0)->setBrush(QBrush(QColor(0, 0, 255, 20))); // first graph will be filled with translucent blue

    QVector<double> x;
    for (int i=0; i<values.size(); ++i)
    {
      x.push_back( i*FREQUENCY_STEP );
    }
    // configure right and top axis to show ticks but no labels:
    // (see QCPAxisRect::setupFullAxesBox for a quicker method to do this)
    ui->plot->xAxis2->setVisible(true);
    ui->plot->xAxis2->setTickLabels(false);
    ui->plot->yAxis2->setVisible(true);
    ui->plot->yAxis2->setTickLabels(false);
    // make left and bottom axes always transfer their ranges to right and top axes:
    connect(ui->plot->xAxis, SIGNAL(rangeChanged(QCPRange)), ui->plot->xAxis2, SLOT(setRange(QCPRange)));
    //connect(ui->plot->yAxis, SIGNAL(rangeChanged(QCPRange)), ui->plot->yAxis2, SLOT(setRange(QCPRange)));
    // pass data points to graphs:
    ui->plot->graph(0)->setData(x, values);
    // let the ranges scale themselves so graph 0 fits perfectly in the visible area:
    ui->plot->graph(0)->rescaleAxes();

    ui->plot->yAxis->setRange(0, 1000*VOLTAGE);
    ui->plot->xAxis->setRange(0, values.size()*FREQUENCY_STEP);

    // Note: we could have also just called ui->plot->rescaleAxes(); instead
    // Allow user to drag axis ranges with mouse, zoom with mouse wheel and select graphs by clicking:
    ui->plot->setInteractions(QCP::iRangeDrag | QCP::iRangeZoom | QCP::iSelectPlottables);

    ui->plot->replot();
}

void MainWindow::openSerialPort()
{
    //Read settings from registry
    QSettings settings("HKEY_CURRENT_USER\\Software\\Rastro\\Kulaks",
                       QSettings::NativeFormat);
    serial->setPortName(settings.value("Serial/port").toString());
    //Try to open serial device
    if (serial->open(QIODevice::ReadWrite))
    {
        if(     serial->setBaudRate(settings.value("Serial/baudrate").toInt()) &&
                serial->setParity( QSerialPort::NoParity ) &&
                serial->setStopBits( QSerialPort::OneStop ) &&
                serial->setFlowControl( QSerialPort::NoFlowControl )
                )
        {
            ui->statusBar->showMessage("Connected");
            ui->actionConnect->setEnabled(false);
            ui->actionDisconnect->setEnabled(true);
        }
        else
        {
            serial->close();
            QMessageBox::critical(this, tr("Error"), serial->errorString());

            ui->statusBar->showMessage(tr("Open error"));
        }
    }
    else
    {
        QMessageBox::critical(this, tr("Error"), serial->errorString());

        ui->statusBar->showMessage(tr("Configure error"));
    }

    //sendSerialData(QByteArray( QString("Hello from Qt").toLatin1() ));
    sendSerialData( "q" ); //RESET the XMEGA
    this->thread()->msleep(150); //Wait for xmega to reinitiliaze
    //Send the configuration to the hardware
    this->sendDeviceSettings( settings.value("Dev/min").toInt(), settings.value("Dev/max").toInt(), settings.value("Dev/n").toInt() );

}

void MainWindow::closeSerialPort()
{
    serial->close();
    ui->statusBar->showMessage(tr("Disconnected"));

    ui->actionConnect->setEnabled(true);
    ui->actionDisconnect->setEnabled(false);
}

void MainWindow::sendSerialData(QByteArray data)
{
    serial->write(data);
    serial->waitForBytesWritten(2000);
}

//reads data from serial
void MainWindow::readData()
{   


    //QByteArray data = serial->readLine();

    QByteArray data = serial->readAll(); //Read all buffer

    if(terminal->isVisible())
        terminal->writeToTerminal(QString(data)); //If terminal is visible print results

    allData.append(data); //Append to our own buffer
    //allData = data;

    //qDebug() << allData.size();
    if(allData.contains("{")) //If buffer contains beining of data
    { //Make sure that unused data beforehand are deleted
        allData.remove(0, allData.indexOf("{"));

        if(!dataASecondT->isActive())
            dataASecondT->start(1000);
    }

    // If data contains both curly brackets, WE GOT DATA
    if(allData.indexOf("{") == 0 && allData.contains("}"))
    {
        QByteArray tempData = allData;

        //Extract the received data
        tempData = tempData.mid(allData.indexOf("{")+1, allData.indexOf("}")-1);
        allData.remove( 0, allData.indexOf( "}" )+1 );

        QString tempStr(tempData);

        QStringList tempList = tempStr.split(" ");

        QVector<double> numbers;

        //Convert extracted data to voltage
        foreach(QString str, tempList)
        {
            if(!str.isEmpty())
            {
                //numbers.push_back( str.toDouble()*VOLTAGE );
                numbers.push_back( (str.toDouble() * 2.7/ 4096.0f) - 0.135f );
            }
        }

        numberOfSamples = numbers.size();


        writeToLogger( QString("Extracted numbers: %1\tGathered data: %2\n").arg(numbers.size()).arg(gatheredData.size()) );

        plot(numbers); // <- plot the new results

        if(ui->actionDo_Magic->isChecked()) //If SVM classifier is enabled
        {
            doSomeMagic( numbers ); //Do pattern recognition
        }

        if(numberOfDataStillNeeded > 1) //Does SVM learner need more data?
        {
            gatheredData.push_back( numbers );
            if(gatheredData.size()%10 == 0)
                ui->progressBar->setValue(numberOfDataNeeded-numberOfDataStillNeeded);
        }
        else if(numberOfDataStillNeeded == 1) //If enough has gathered, save to SVM file
        {
            ui->progressBar->setVisible(false);
            gatheredData.push_back( numbers );

            //time to save svm file
            QFile file(this->file);
            if (!file.open(QIODevice::Append))
            {
                QMessageBox::information(this, tr("Unable to open file"),
                    file.errorString());
                return;
            }
            else
            {
                QTextStream out(&file);

                if(file.size() > 0)
                    out << endl;

                for(int d=0; d<gatheredData.size(); d++)//QVector< double > data, gatheredData)
                {
                    QVector<double> data = gatheredData.at(d);
                    out << this->id << " ";

                    for(int i=0; i<data.size(); i++)
                        if(data.size()-1 != i || d == gatheredData.size()-1)
                            out << QString::number(i+1) << ":" << QString::number( data[i] ) << " ";
                        else
                            out << QString::number(i+1) << ":" << QString::number( data[i] ) << endl;
                }

                writeToLogger("Data saved to " + this->file);
                file.close();
            }

            //time to save name file
            this->file.append(".name");
            file.setFileName(this->file);
            if (!file.open(QIODevice::Append))
            {
                QMessageBox::information(this, tr("Unable to open file"),
                    file.errorString());
                return;
            }
            else
            {
                QTextStream out(&file);
                out << this->gestureName << endl;
            }

            file.close();
        }


        if(numberOfDataStillNeeded != 0)
            numberOfDataStillNeeded--;

        collected++;
    }


}

void MainWindow::dataASecond()
{
    //qDebug() << collected;
    ui->lblDataASecond->setText(QString::number(collected) + " data/second");

    //Code for measuring the performance of the device.
    //Every second stores how many data points were collected
//    QFile file ("avg_conv_data.txt");
//    if (!file.open(QIODevice::Append))
//    {
//        QMessageBox::information(this, tr("Unable to open file"),
//            file.errorString());
//        return;
//    }
//    else
//    {
//        QTextStream out(&file);
//        out << QString::number(collected) << " ";
//    }

//    file.close();

    collected = 0;
}

void MainWindow::writeToLogger(QString text)
{
    //Writes text to bottom debugger text area

    if(ui->console->toPlainText().length() > 10000)
        ui->console->clear();
    ui->console->insertPlainText( text );
    QTextCursor c =  ui->console->textCursor();
    c.movePosition(QTextCursor::End);
    ui->console->setTextCursor(c);

}

MainWindow::~MainWindow()
{
    delete ui;
}
